/*
 * Copyright (C) 1997-2005 Andre Noll <maan@systemlinux.org>
 *
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 *
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.
 */


#include "server.h"

/*
 * time (in milliseconds) that has to pass between death of stream
 * writer and the start of the next stream writer. Mainly to cope with
 * user errors such as an invalid stream write command that would make
 * para_server hog the CPU without that delay.
 */
#define SW_RESTART_TIME 100
static struct timeval sw_restart_tv;
static struct timeval sw_restart_barrier;
/*
 * Delay betweeen announcing the stream via the stat command and
 * sending the first byte.
 */
#define ANNOUNCE_TIME 100
static struct timeval announce_tv;
static struct timeval data_send_barrier;
/* 
 * Some streamers (atm only ogg) like to send the stream in a slightly
 * shorter time than the actual playing time of the file. The amount of
 * time may depend on the file being streamed and has to be corrected
 * after closing the stream writer fd. We let the audio plugin write
 * the "real" stream end to this variable and won't try to get the next 
 * song before eof_barrier.
 */
long unsigned eof_time;
static struct timeval eof_barrier, eof_tv;

/* the time when the last stream was started/died */
static struct timeval sw_exec, sw_death;

extern struct misc_meta_data *mmd;
extern struct dbtool dblist[];

static FILE *audio_file = NULL;
static char supported_audio_formats[MAXLINE];

#if 1
	int mp3_init(void *);
#endif

#ifdef HAVE_OGGVORBIS
	int ogg_init(void *);
#endif

struct audio_format afl[] = { /* audio format list */
#if 1
	{
		.name = "mp3",
		.init = mp3_init,
		.sw_pid = -1,
		.fifo = -1,
		.open_cmd = NULL,
	},
#endif
#ifdef HAVE_OGGVORBIS
	{
		.name = "ogg",
		.init = ogg_init,
		.sw_pid = -1,
		.fifo = -1,
		.open_cmd = NULL,
	},
#endif
	{
		.name = NULL,
	}
};

const char *audio_format_name(int i)
{
	return i >= 0?  afl[i].name : "(none)";
}

char *afs_stream_writer_cmd(int i)
{
	return i >= 0? afl[i].open_cmd : "/bin/false";
}


char *afs_supported_audio_formats(void)
{
	return supported_audio_formats;
}

void afs_handle_sigchild(pid_t pid)
{
	int i;

	if (pid <= 0) {
		clog(CRIT, "%s: BUG: pid == %d\n", __func__, pid);
		return;
	}
	for (i = 0; afl[i].name; i++) {
		if (afl[i].sw_pid == pid) {
			clog(INFO, "%s: %s stream writer (pid %d) died\n",
				__func__, afl[i].name, pid);
			afl[i].sw_pid = -1;
			mmd->new_afs_status_flags &= ~AFS_STREAM_WRITER;
			mmd->chunks_sent = 0;
			gettimeofday(&sw_death, NULL);
			tv_add(&sw_death, &sw_restart_tv, &sw_restart_barrier);
			ms2tv(eof_time, &eof_tv);
			tv_add(&eof_tv, &sw_death, &eof_barrier);
		}
	}
}

/*
 * Although the fifo is only used for writing we open it for reading as
 * well. That prevents an eof beeing generated if the number of clients
 * goes from 1 to 0 (ancient trick described in APUE).
 *
 * Set close on exec flag for both fds to mark it for closing after
 * fork. Otherwise closing the pipe does not have the desired effect as
 * the child of the server still holds open the fifo.
 */
int afs_open_fifo(char *fifoname, int fd[2])
{
	int flags;

	fd[0] = para_open_fifo(fifoname, O_RDONLY | O_NONBLOCK);
	if (fd[0] < 0) {
		clog(CRIT, "%s: read fifo open error (%d)\n", __func__, fd[0]);
		return -1;
	}
	fd[1] = para_open_fifo(fifoname, O_WRONLY | O_NONBLOCK);
	if (fd[1] < 0) {
		clog(CRIT, "%s: write fifo open error (%d)\n", __func__, fd[1]);
		return -2;
	}
	flags = fcntl(fd[1], F_GETFL);
	fcntl(fd[1], F_SETFL, flags & ~ O_NONBLOCK);
	fcntl(fd[1], F_SETFD, FD_CLOEXEC);
	fcntl(fd[0], F_SETFD, FD_CLOEXEC);
//	close(r);
	clog(DEBUG, "%s: %s opened (readfd: %d, writefd: %d)\n", __func__,
		fifoname, fd[0], fd[1]);
	mmd->events++;
	return 1;
}

static int setup_stream_command(struct gengetopt_args_info *conf, int num)
{
	int ret, i;
	struct audio_format *af = &afl[num];
	const char *n = af->name;
	int g = conf->stream_write_cmd_given;
	char *hn = para_hostname(), *home = para_homedir();

	free(af->open_cmd);
	af->open_cmd = NULL;
	ret = 0;
	if (!g)
		goto out;
	clog(DEBUG, "%s: %d stream write command%s given, checking for %s\n",
		__func__, g, g > 1? "s" : "", n);
	for (i = g - 1; i >= 0; i--) {
		char *tmp, *c = conf->stream_write_cmd_arg[i];
		size_t len = strlen(n);
		if (strlen(c) <= len || strncmp(c, n, len))
			continue;
		if (*(c + len) != ':') {
			clog(WARNING, "%s: invalid stream_write_cmd\n", __func__);
			ret = -1;
			goto out;
		}
		af->open_cmd = para_strdup(c + len + 1);
		tmp = s_a_r(af->open_cmd, "HOME", home);
		af->open_cmd = s_a_r(tmp, "HOSTNAME", hn);
		free(tmp);
		clog(INFO, "%s: using \"%s\" for files of type %s\n",
			__func__, af->open_cmd, n);
		ret = 1;
		goto out;
	}
out:
	free(hn);
	free(home);
	return ret;
}

int afs_handle_sighup(struct gengetopt_args_info *conf)
{
	int i, ret;
	for (i = 0; afl[i].name; i++)
		if ((ret = setup_stream_command(conf, i)) < 0)
			return ret;
	return 1;
}

int afs_init(struct gengetopt_args_info *conf)
{
	int i, ret;
	char *hn = para_hostname(), *home = para_homedir();

	clog(INFO, "%s: checking audio formats\n", __func__);
	if (!home)
		goto out;
	*supported_audio_formats = '\0';
	for (i = 0; afl[i].name; i++) {
		clog(DEBUG, "%s: supported audio format: %s\n", __func__,
			afl[i].name);
		if ((ret = setup_stream_command(conf, i)) < 0)
			goto out;
		afl[i].fifoname = make_message("%s/.paraslash/server_%s_fifo.%s",
			home, afl[i].name, hn);
		if (afl[i].init(&afl[i]) < 0) {
			clog(ERROR, "%s: init %s failed, disabling "
				"%s support\n", __func__, afl[i].name,
				afl[i].name);
			afl[i].get_file_info = NULL;
			continue;
		}
		sprintf(supported_audio_formats, "%s%s ",
			*supported_audio_formats? supported_audio_formats: "",
			 afl[i].name);
	}
	ret = -1;
	if (!i) {
		clog(EMERG, "%s: no audio formats supported\n", __func__);
		goto out;
	}
	clog(DEBUG, "%s: supported audio formats: %s\n", __func__,
		supported_audio_formats);
	ret = i;
	ms2tv(ANNOUNCE_TIME, &announce_tv);
	ms2tv(SW_RESTART_TIME, &sw_restart_tv);
out:
	free(hn);
	free(home);
	return ret;
}

static int get_file_info(int i)
{
	/* get_file_info must not close the audio file, even if it is invalid */
	int ret = afl[i].get_file_info(audio_file, mmd->audio_file_info,
		&mmd->chunks_total, &mmd->seconds_total);
	if (ret > 0)
		clog(INFO, "%s: valid %s file\n", __func__, afl[i].name);
	else
		clog(DEBUG, "%s: %s format not recognized\n", __func__,
			afl[i].name);
	return ret;
}

static int guess_audio_format(const char *name)
{

	int i, len1 = strlen(name), len2;

	for (i = 0; afl[i].name; i++) {
		len2 = strlen(afl[i].name);
		if (len1 < len2 || !afl[i].get_file_info)
			continue;
		if (!strncasecmp(name + (len1 - len2), afl[i].name, len2)) {
			clog(DEBUG, "%s: might be %s\n", __func__,
				afl[i].name);
			return i;
		}
	}
	return -1;
}

static int get_audio_format(int omit)
{
	int i;

	for (i = 0; afl[i].name; i++) {
		if (i == omit || !afl[i].get_file_info)
			continue;
		rewind(audio_file);
		if (get_file_info(i) > 0)
			return i;
		rewind(audio_file);
	}
	return -1;
}

/*
 * upddate shared mem
 */
static int update_mmd(void)
{
	int i;
	struct stat file_status;

	i = guess_audio_format(mmd->filename);
	if (i < 0 || get_file_info(i) < 0)
		i = get_audio_format(i);
	if (i < 0) {
		clog(WARNING, "%s: unknown audio format for %s\n", __func__,
			mmd->filename);
		return -1;
	}
	mmd->audio_format = i;
	mmd->chunks_sent = 0;
	mmd->current_chunk = 0;
	mmd->offset = 0;
	if (fstat(fileno(audio_file), &file_status) == -1) {
		clog(ERROR, "%s: fstat error (%s)\n", __func__, strerror(errno));
		return -1;
	}
	mmd->size = file_status.st_size;
	mmd->mtime = file_status.st_mtime;
	mmd->events++;
	clog(NOTICE, "%s: next audio file: %s\n", __func__, mmd->filename);
	return 1;
}

static void get_song(void)
{
	char **sl = NULL;
	int i;

	if (mmd->new_afs_status_flags & AFS_NOMORE)
		goto err_out;

	sl = dblist[mmd->dbt_num].get_song_list(10);
	if (!sl)
		goto err_out;
	for (i = 0; sl[i]; i++) {
		clog(INFO, "%s: trying %s\n", __func__, sl[i]);
		if (strlen(sl[i]) >= _POSIX_PATH_MAX)
			continue;
		audio_file = fopen(sl[i], "r");
		if (!audio_file)
			continue;
		strcpy(mmd->filename, sl[i]);
		if (update_mmd() < 0) {
			fclose(audio_file);
			continue;
		}
		mmd->num_played++;
		dblist[mmd->dbt_num].update_song(sl[i]);
		clog(DEBUG, "%s: success\n", __func__);
		mmd->new_afs_status_flags &= (~AFS_NEXT);
		goto free;
	}
	clog(ERROR, "%s: no valid files found\n", __func__);
err_out:
	mmd->new_afs_status_flags = AFS_NEXT;
free:
	if (sl) {
		for (i = 0; sl[i]; i++)
			free(sl[i]);
		free(sl);
	}
}

int call_all_mainloops(struct timeval **next_time)
{
	struct timeval *dummy;
	int i, fd = -1, format = mmd->audio_format;
	unsigned int flags = mmd->new_afs_status_flags;

	*next_time = NULL;
	for (i = 0; afl[i].name; i++) {
//	clog(DEBUG, "%s: i = %d, format = %d\n", __func__, i, format);
		if (i == format) {
			fd = afl[i].mainloop(1, flags, next_time, 
				mmd->current_chunk, mmd->chunks_sent, 
				mmd->repos_request, &mmd->offset,
				&mmd->stream_start, &sw_exec,  &eof_time);
			tv_add(&sw_exec, &announce_tv, &data_send_barrier);
		} else
			afl[i].mainloop(0, flags, &dummy, mmd->current_chunk, 
				mmd->chunks_sent, mmd->repos_request, &mmd->offset,
				&mmd->stream_start, &sw_exec, &eof_time);
		if (afl[i].sw_pid > 0)
			mmd->new_afs_status_flags |= AFS_STREAM_WRITER;
	}
	return fd;
}

static int chk_barrier(const char *bname, const struct timeval *now, 
		const struct timeval *barrier, struct timeval *diff, int log)
{
	long ms;

	if (tv_diff(now, barrier, diff) > 0)
		return 1;
	ms = tv2ms(diff);
	if (log && ms)
		clog(DEBUG, "%s: %s barrier: %lims left\n", __func__, bname,
			ms);
	return -1;
}

/*
 * < 0: time left. remainder was written to timeout
 * > 0: no next_time or next_time expired
 */
static int afs_compute_timeout(struct timeval *next_time, struct timeval **timeout)
{
	static struct timeval the_timeout;
	struct timeval now;

	*timeout = &the_timeout;
	gettimeofday(&now, NULL);
	if (chk_barrier("eof", &now, &eof_barrier, &the_timeout, 1) < 0)
		return -1;
	if (chk_barrier("data send", &now, &data_send_barrier,
			 &the_timeout, 1) < 0)
		return -1;
	if (chk_barrier("sw restart", &now, &sw_restart_barrier,
		 &the_timeout, 1) < 0)
		return -1;
	if (!next_time || mmd->audio_format < 0)
		return 1;
	if (chk_barrier(afl[mmd->audio_format].name, &now, next_time,
			&the_timeout, 0) < 0)
		return -1;
	return 1;
}

/*
 * called from main in any case, just before main calls select
 */
int afs_mainloop(struct timeval **timeout)
{
	struct audio_format *af = NULL;
	struct timeval *next_time;
	int ret, sw_fileno, playing, next, format, repos;

	*timeout = NULL;
again:
	playing = mmd->new_afs_status_flags & AFS_PLAYING;
	next = mmd->new_afs_status_flags & AFS_NEXT;
	format = mmd->audio_format;
	repos = mmd->new_afs_status_flags & AFS_REPOS;
	if (format >= 0)
		af = afl + format;
	ret = -1;
	sw_fileno = call_all_mainloops(&next_time);
	if (repos) {
		mmd->new_afs_status_flags &= ~(AFS_REPOS);
		mmd->current_chunk = mmd->repos_request;
	}
	if (next) {
		audio_file = NULL;
		mmd->audio_format = -1;
		/* do nothing until stream writer dies */
		if (mmd->new_afs_status_flags & AFS_STREAM_WRITER)
			goto out;
	}
get_audio_file:
	ret = afs_compute_timeout(next_time, timeout);
	if (ret < 0)
		goto out;
	if (!audio_file && playing) {
		clog(DEBUG, "%s: ready and playing, but no audio file\n",
			__func__);
		get_song();
		goto again;
	}
//	clog(DEBUG, "%s: pid: %i\n", __func__, af? af->sw_pid : -42);
	ret = -1;
	if (!playing || !audio_file || !af || af->sw_pid < 0)
		goto out;
	if (ret > 0 && !audio_file) {
		next_time = NULL;
		goto get_audio_file;
	}
	ret = sw_fileno;
out:
//	clog(DEBUG, "%s: returning fd %d, tv = %lims\n", __func__,
//		ret, *timeout? tv2ms(*timeout) : 0);

	return ret;
}

/*
 * Status is playing and stream writer is ready for writing.
 */
int afs_send_chunk(void)
{
	int ret;
	struct audio_format *af;

	if (mmd->audio_format < 0 || !audio_file) {
		clog(CRIT, "%s: BUG\n", __func__);
		return -1;
	}
	if (!mmd->chunks_sent) {
		gettimeofday(&mmd->stream_start, NULL);
		mmd->events++;
	}
	af = &afl[mmd->audio_format];
	ret = af->send_chunk(mmd->audio_file_info, mmd->current_chunk,
		mmd->chunks_sent, mmd->new_afs_status_flags);
	mmd->new_afs_status_flags &= ~(AFS_NEXT | AFS_REPOS);
	if (ret > 0) {
		mmd->new_afs_status_flags |= AFS_PLAYING;
		mmd->chunks_sent++;
		mmd->current_chunk++;
	}
	if (ret < 0)
		mmd->new_afs_status_flags = AFS_NEXT;
	if (!ret) {
		struct timeval now, stream_time;
		gettimeofday(&now, NULL);
		tv_diff(&now, &mmd->stream_start, &stream_time);
		clog(INFO, "%s: eof, sent: %d/%d. "
			"stream time: %d:%d\n",
			__func__, mmd->current_chunk, mmd->chunks_total,
			stream_time.tv_sec, stream_time.tv_usec / 1000
		);
		mmd->new_afs_status_flags |= AFS_NEXT;
	}
	if (ret <= 0) {
		mmd->audio_format = -1;
		mmd->events++;
	}
	return ret;
}
